import express = require('express');
import {Request, Response, NextFunction} from 'express';
declare module 'express' {
    interface Request {
        log?: any,
        id?: string
    }
}
import http = require('http');
import cookieParser = require('cookie-parser');
import bodyParser = require('body-parser');
import serveStatic = require('serve-static');
import initRoutes from '../routes';
import sentry from './sentry';
import pino from './pino';

// const Arena: any = require('bull-arena');
const uuidv4: () => string = require('uuid/v4');

const devMode = process.env.APP_ENV === 'development';

const httpLogger: any = require('pino-http')({
    logger: pino,
    genReqId: () => uuidv4(),
    serializers: {
        req: (req: Request) => {
            if (devMode) {
                return `${req.method} ${req.url} [${req.id}]`;
            } else {
                return {
                    id: req.id,
                    method: req.method,
                    url: req.url,
                    headers: req.headers,
                    remoteAddress: req.connection && req.connection.remoteAddress,
                    remotePort: req.connection && req.connection.remotePort
                };
            }
        },
        res: (res: Response) => {
            if (devMode) {
                const cleanedHeaders = String(res.header)
                    .replace(/(\r\n)+/g, ' • ')
                    .replace(/\s•\s$/, '');
                return `[${res.statusCode}] ${cleanedHeaders}`;
            } else {
                return {
                    statusCode: res.statusCode,
                    header: res.header
                };
            }
        }
    },
});

export default function generateServer() {
    const app = express();
    const router = express.Router(); // eslint-disable-line new-cap

    app.disable('x-powered-by');
    app.use(httpLogger);
    app.use(sentry.requestHandler());
    app.use(bodyParser.json());
    app.use(bodyParser.urlencoded({extended: true}));
    app.use(cookieParser());

    if (devMode) {
        app.use(serveStatic(`${__dirname}/../../public`));
    }

    initRoutes(router);
    // app.use('/', Arena({
    //     queues: [
    //         {
    //             name: 'common',
    //             hostId: 'scpdb.org',
    //             type: 'bee'
    //         }
    //     ]
    // }, {disableListen: true, basePath: '/o5/queue'}));
    app.use('/', router);

    app.use(sentry.errorHandler());

    app.use((err: Error, req: Request, res: Response, next: NextFunction) => { // eslint-disable-line no-unused-vars
        if (req.log) {
            req.log.error(err, 'Unhandled error')
        } else {
            pino.error(err, 'Unhandled error');
        }
        res.end('Internal server error');
    });

    return function listen(port: number) {
        const server = http.createServer(app);
        server.on('listening', () => {
            pino.info(`Express server is listening on port ${server.address().port}`);
        });
        server.on('error', (error) => {
            if ((error as any).syscall !== 'listen') {
                throw error;
            }
            switch ((error as any).code) {
                case 'EACCES':
                    pino.error(`Port ${port} requires elevated privileges`);
                    process.exit(1);
                    break;
                case 'EADDRINUSE':
                    pino.error(`Port ${port} is already in use`);
                    process.exit(1);
                    break;
                default:
                    throw error;
            }
        });
        server.listen(port);
    }
}
